@@ -5,21 +5,23 @@ const app = angular.module('dataiku.agenttools', ['dataiku.services', 'dataiku.f
 
 app.controller('AgentToolController', function ($scope, $state, $stateParams, $q, $timeout, WT1, DataikuAPI, PromptUtils, TopNav, DatasetUtils, FutureProgressModal,
     $rootScope, FutureWatcher, Debounce, CreateModalFromTemplate, TaggingService, Dialogs, localStorageService, SmartId, translate, ClipboardUtils, AgentToolService,
-    SavedModelsService, PluginsService, AgentToolCodeTemplates, ActiveProjectKey, Expressions, CodeBasedEditorUtils, DKUtils) {
+    SavedModelsService, PluginsService, AgentToolCodeTemplates, ActiveProjectKey, Expressions, CodeBasedEditorUtils, DKUtils, RetrievableKnowledgeUtils,
+    CodeMirrorSettingService) {
     $scope.uiState = {};
 
+    // inline python tools
     $scope.editorOptions = CodeBasedEditorUtils.editorOptions('text/x-python', $scope, true);
     CodeBasedEditorUtils.registerBroadcastSelectionHandler($scope.editorOptions);
     CodeBasedEditorUtils.setRecipeScript($scope.script);
 
+    // local MCP tools
+    $scope.codeMirrorSettingService = CodeMirrorSettingService;
+
     let savedAgentTool;
     $scope.set = function(agentTool) {
         $scope.agentToolTypeUiData = AgentToolService.getAgentToolTypesUiData(agentTool);
         $scope.agentTool = agentTool;
         savedAgentTool = angular.copy(agentTool);
-        $timeout(function() {
-            $scope.$broadcast('agentToolSetEvent');
-        }, 0);
     };
 
     $scope.isDirty = function() {
@@ -32,6 +34,9 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
         if (!$scope.isDirty()) {
             return $q.when('agent tool not dirty');
         }
+        if ($scope.isDisabled && $scope.isDisabled()) {
+            return $q.when('invalid agent tool state');
+        }
 
         const deferred = $q.defer();
         const saveAfterConflictCheck = function() {
@@ -44,7 +49,7 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                 savedAgentTool = angular.copy($scope.agentTool);
                 $scope.$broadcast('agentToolSavedEvent');
                 deferred.resolve('agent tool saved');
-            }, () => deferred.reject())
+            }, () => deferred.reject()).error(setErrorInScope.bind($scope));
         };
 
         DataikuAPI.agentTools.checkSaveConflict($scope.agentTool).success(function(conflictResult) {
@@ -71,6 +76,9 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
         if (!$scope.isDirty()) {
             return $q.when('agent tool not dirty');
         }
+        if ($scope.isDisabled && $scope.isDisabled()) {
+            return $q.when('invalid agent tool state');
+        }
 
         let frankenTool = angular.copy($scope.agentTool);
         frankenTool["quickTestQuery"] = savedAgentTool["quickTestQuery"];
@@ -93,16 +101,264 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
 
     $scope.showToolDescriptionModal = function(agentTool){
         DataikuAPI.agentTools.getDescriptor(agentTool.projectKey, agentTool.id).success(descriptor => {
-            $scope.toolDescriptionModal = {
-                // remove the global description that was appended by the back-end in get-descriptor to show it in another field
-                internalDescription: descriptor.description.replace("\n\n" + agentTool.additionalDescriptionForLLM, ""),
-                globalDescription: agentTool.additionalDescriptionForLLM,
-                agentSpecificDescription: undefined,
-            }
-            CreateModalFromTemplate('/templates/savedmodels/agents/agent-tool-description-modal.html', $scope);
+            const modalScope = $scope.$new(true);
+            modalScope.toolName = agentTool.name;
+            // remove the global description that was appended by the back-end in get-descriptor to show it in another field
+            modalScope.internalDescription = descriptor.description.replace(agentTool.additionalDescriptionForLLM, "").trimEnd();
+            modalScope.globalDescription = agentTool.additionalDescriptionForLLM;
+            CreateModalFromTemplate('/templates/savedmodels/agents/agent-tool-description-modal.html', modalScope);
         }).error(setErrorInScope.bind($scope));
     }
 
+    $scope.isToolRunWithKernelPool = function() {
+        const toolsRunWithKernelPool = ["VectorStoreSearch", "InlinePython", "GenericStdioMCPClient"];
+        return ($scope.agentTool.type.indexOf("Custom") == 0) || toolsRunWithKernelPool.includes($scope.agentTool.type);
+    };
+
+    //
+    // Start of local MCP tool specific
+    //
+    $scope.showMCPJsonConfigModal = function(){
+        CreateModalFromTemplate('/templates/agent-tools/agent-tool-mcp-json-modal.html', $scope);
+    }
+
+    $scope.subtools = undefined;
+    $scope.loadSubtools = function() {
+        if (!$scope.agentTool || $scope.agentTool.type !== "GenericStdioMCPClient" || !$scope.agentTool.params) return;
+        $scope.subtools = undefined;
+        $scope.subtoolsError = undefined;
+        $scope.loadingSubtools = true;
+
+        const loadDescriptor = function() {
+            DataikuAPI.agentTools.getDescriptor($stateParams.projectKey, $stateParams.agentToolId).success(data => {
+                $scope.subtools = data.subtools;
+                $scope.subtoolsError = undefined;
+                $scope.loadingSubtools = false;
+            }).error((data, status, headers, _, statusText) => {
+                $scope.subtools = undefined;
+                $scope.loadingSubtools = false;
+                if (data instanceof Error) {
+                    throw data;
+                }
+                const apiError = getErrorDetails(data, status, headers, statusText);
+                $scope.subtoolsError = apiError;
+
+                /* Report to WT1 */
+                window.APIErrorLogger.error("API error", apiError);
+                window.DKUErrorReporting.reportBackendAPIError(apiError);
+            });
+        }
+
+        $scope.saveIgnoringQuickTestQuery().then(loadDescriptor);
+    }
+
+
+    $scope.selection = {};
+
+    $scope.areAllSelectedSubtoolsActive = function() {
+        if (!$scope.selection || !$scope.selection.selectedObjects) {
+            return false;
+        }
+        return $scope.selection.selectedObjects.every(function(tool) {
+            return tool.enabled;
+        });
+    };
+
+    $scope.enableSelectedSubtools = function() {
+        $scope.selection.selectedObjects.map(subtool => {
+            $scope.toggleSubtool(subtool, true);
+        });
+    };
+
+    $scope.disableSelectedSubtools = function() {
+        $scope.selection.selectedObjects.map(subtool => {
+            $scope.toggleSubtool(subtool, false);
+        });
+    };
+
+    $scope.getSubtoolToggleState = function(tool) {
+        if ($scope.agentTool.params.subtoolsEnabledByDefault) {
+            return true;
+        }
+        if (tool.name in $scope.agentTool.params.subtoolsStateOverride) {
+            return $scope.agentTool.params.subtoolsStateOverride[tool.name];
+        }
+        return false;
+    };
+
+    $scope.toggleSubtool = function(tool, state) {
+        if (!$scope.agentTool || $scope.agentTool.type !== "GenericStdioMCPClient" || !$scope.agentTool.params) return;
+
+        if (state !== undefined) {
+            tool.enabled = state;
+            $scope.agentTool.params.subtoolsStateOverride[tool.name] = state;
+        } else {
+            if (tool.name in $scope.agentTool.params.subtoolsStateOverride) {
+                $scope.agentTool.params.subtoolsStateOverride[tool.name] = !$scope.agentTool.params.subtoolsStateOverride[tool.name];
+            } else {
+                $scope.agentTool.params.subtoolsStateOverride[tool.name] = !$scope.agentTool.params.subtoolsEnabledByDefault;
+            }
+        }
+    };
+
+    $scope.subtoolsEnabledByDefaultChanged = function(newVal) {
+        // nothing to do when it's switched back ON
+        if (newVal) {
+            return;
+        }
+        // mark every unknown tool as enabled when it's switched OFF
+        $scope.subtools?.map(subtool => {
+           if (!(subtool.name in $scope.agentTool.params.subtoolsStateOverride)) {
+               $scope.toggleSubtool(subtool, true);
+           }
+        });
+    }
+
+    function toolSchemaProperty(property, required) {
+        // example input property
+        // {
+        //     "type": "object",
+        //     "properties": {
+        //         "from": {
+        //             "type": "string",
+        //             "description": "The name of the entity"
+        //         },
+        //     },
+        //     "required": ["from"]
+        // }
+        // example return values
+        // {'from': '(required) (string) The name of the entity'}
+        // "(required) (number) The public ID of the story"
+        // "(string) The type of relationship (one of 'relates to', 'blocks', 'blocked by', 'duplicates', 'duplicated by')"
+        if (property === undefined) {
+            return "";
+        }
+
+        let str = required ? "(required)" : "";
+
+        if (property.type !== undefined) {
+            if (str !== "") {
+                str += " ";
+            }
+
+            if (property.type === "array") {
+                if (property.items && property.items.type) {
+                    if (property.items.type === "object" || property.items.type === "array") {
+                        return Array(toolSchemaProperty(property.items, null));
+                    } else {
+                        str += "(array of " + property.items.type + ")";
+                    }
+                } else {
+                    str += "(array)";
+                }
+            } else if (property.type === "object") {
+                let obj = {}
+                for (const name in property.properties) {
+                    obj[name] = toolSchemaProperty(property.properties[name], property.required && property.required.includes(name));
+                }
+                return obj;
+            } else {
+                str += "(" + property.type + ")";
+            }
+        }
+        if (property.description !== undefined) {
+            if (str !== "") {
+                str += " ";
+            }
+            str += property.description;
+        }
+        if (property.enum !== undefined) {
+            if (str !== "") {
+                str += " ";
+            }
+            str += "(one of '" + property.enum.join("', '") + "')";
+        }
+        return str;
+    }
+
+    function getTestQueryInput(tool) {
+        if (!tool.inputSchema) {
+            return {};
+        }
+        return toolSchemaProperty(tool.inputSchema, null);
+    }
+
+    $scope.prefillTestQuery = function(tool) {
+        $scope.agentTool.quickTestQuery = {
+            "subtoolName": tool.name,
+            "input": getTestQueryInput(tool),
+            "context": {}
+        };
+    }
+
+    $scope.validateJSONConfig = function(config) {
+        $scope.overwriteServerSettingsError = null;
+        if (!config || typeof config !== 'object' || Array.isArray(config)) {
+            $scope.overwriteServerSettingsError = "Invalid configuration, it should be a JSON object";
+            return;
+        }
+
+        // extract the server if it's nested in { mcpServers: { serverName: { command, args, env } } }
+        if (config && config.mcpServers != null && typeof config.mcpServers === 'object') {
+            const serverKeys = Object.keys(config.mcpServers);
+            if (serverKeys.length > 0) {
+                config = config.mcpServers[serverKeys[0]];
+            }
+        }
+        // extract the server if it's nested in { servers: { serverName: { command, args, env } } }
+        else if (config && config.servers != null && typeof config.servers === 'object') {
+            const serverKeys = Object.keys(config.servers);
+            if (serverKeys.length > 0) {
+                config = config.servers[serverKeys[0]];
+            }
+        }
+        // extract the server if it's nested in { serverName: { command, args, env } }
+        else if (config && config.command == null) {
+            const serverKeys = Object.keys(config);
+            if (serverKeys.length > 0) {
+                config = config[serverKeys[0]];
+            }
+        }
+
+        if (!config.command) {
+            $scope.overwriteServerSettingsError = 'Failed to parse MCP server configuration, it should be a valid Claude desktop configuration and have a "command" field';
+            return;
+        }
+
+        return config;
+    }
+
+    $scope.overwriteServerSettingsFromJSON = function(payload) {
+        const config = $scope.validateJSONConfig(payload);
+        if (!config) {
+            return;
+        }
+
+        $scope.agentTool.params.command = config.command;
+
+        let args = [];
+        if (config.args !== null && Array.isArray(config.args)) {
+            args = config.args;
+        }
+        $scope.agentTool.params.args = args;
+
+        let env = [];
+        if (config.env !== null && typeof config.env === 'object') {
+            for (const key in config.env) {
+                if (config.env.hasOwnProperty(key)) {
+                    env.push({ name: key, value: config.env[key], secret: false });
+                }
+            }
+        }
+        $scope.agentTool.params.env = env;
+
+        return true;
+    }
+
+    //
+    // End of local MCP tool specific
+    //
+
     function load() {
         DatasetUtils.listDatasetsUsabilityForAny($stateParams.projectKey).success(data => {
             $scope.availableDatasets = data;
@@ -128,19 +384,15 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
 
             if ($scope.agentTool.type == 'VectorStoreSearch') {
                 // keep in sync with retrievable-knowledge.js
-                let kbDefaults = {
-                    maxDocuments: 4,
-                    searchType: "SIMILARITY",
-                    similarityThreshold: 0.5,
-                    mmrK: 20,
-                    mmrDiversity: .25,
-                    useAdvancedReranking: false,
-                    rrfRankConstant: 60,
-                    rrfRankWindowSize: 4,
-                };
+                let kbDefaults = RetrievableKnowledgeUtils.getRAGLLMSettings();
+                const dku_embedding_column_option = {
+                    value:"DKU_TEXT_EMBEDDING_COLUMN",
+                    label: "Embedding column (Always included)",
+                    disabled: true
+                }
                 if ($scope.agentTool.params === undefined) $scope.agentTool.params = [];
                 Object.keys(kbDefaults).forEach((field) => {
-                    if (!$scope.agentTool.params[field])
+                    if ($scope.agentTool.params[field] === undefined)
                         $scope.agentTool.params[field] = kbDefaults[field];
                 });
 
@@ -148,10 +400,7 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                     "columns" : []
                 };
                 $scope.knowledgeBankSchemaColumnNames = [];
-
-                $scope.getSelectableKBColumns = function() {
-                    return angular.copy($scope.knowledgeBankSchemaColumnNames);
-                };
+                $scope.retrievalColumnsOptions = [dku_embedding_column_option]
 
                 $scope.$watch("agentTool.params.knowledgeBankRef", Debounce().withScope($scope).withDelay(50,200).wrap(function(){
                     if (!($scope.agentTool && $scope.agentTool.params && $scope.agentTool.params.knowledgeBankRef)) return;
@@ -160,6 +409,7 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                         $scope.knowledgeBank = retrievableKnowledge;
                         $scope.knowledgeBankSchema.columns =  $scope.knowledgeBank.metadataColumnsSchema || [];
                         $scope.knowledgeBankSchemaColumnNames = $scope.knowledgeBankSchema.columns.map((column) => column.name);
+                        $scope.retrievalColumnsOptions = $scope.retrievalColumnsOptions.concat($scope.knowledgeBankSchemaColumnNames.map(column => {return {value: column, label: column, disabled: false}}))
                     }).error(setErrorInScope.bind($scope));
 
                     DataikuAPI.retrievableknowledge.getFullInfo(ActiveProjectKey.get(), SmartId.create($scope.agentTool.params.knowledgeBankRef, $stateParams.projectKey)).then(function({data}){
@@ -198,6 +448,7 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
             }
 
             if ($scope.agentTool.type == "DataikuReporter") {
+                // see ReporterAgentTool.java
                 $scope.integrationTypes = [
                     { "id" : "slack-direct", "label" : translate("SCENARIO.INTEGRATIONS.TYPES.SLACK", "Slack") },
                     { "id" : "msft-teams-direct", "label" : translate("SCENARIO.INTEGRATIONS.TYPES.TEAMS", "Microsoft Teams") },
@@ -212,6 +463,16 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                 if (!$scope.agentTool.params.variables) {
                     $scope.agentTool.params.variables = []
                 }
+                if (!$scope.agentTool.params.integration.type) {
+                    $scope.agentTool.params.integration.type = "mail-direct";
+                }
+                if (!$scope.agentTool.params.integration.configuration) {
+                    $scope.agentTool.params.integration.configuration = {};
+                }
+                $scope.fileTemplateVariables = [];
+                $scope.undefinedTemplateVariables = new Set();
+                $scope.unusedToolVariables = new Set();
+
                 const fetchDssVariables = function () {
                     const promise = DataikuAPI.flow.recipes.generic.getVariablesWithProjectKey($scope.agentTool.projectKey);
                     promise.success(function(data) {
@@ -237,7 +498,111 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                             }
                         };
                     }, true, 'static', true);
+                };
+
+                const extractVariables = function(text) {
+                    const pattern = /\${([^${}]+)}/g;
+                    return [...text.matchAll(pattern)].map(match => match[1].trim()).filter(match => match !== "");
+                };
+
+                const loadEmailTemplate = function() {
+                    let templateName = $scope.agentTool.params.integration.configuration.templateName  || "default.ftl";
+                    DataikuAPI.agentTools.getEmailTemplateContent(templateName).success(function(data){
+                        $scope.fileTemplateVariables = extractVariables(data.templateContent);
+                        compareTemplateAndToolVariables();
+                    }).error(setErrorInScope.bind($scope));
                 }
+
+                $scope.isDisabled = function() {
+                    return $scope.agentTool.params.variables.some((variable) => variable.$invalid);
+                }
+
+                const getTemplateVariablesSet = function() {
+                    const templateFields = {
+                        "mail-direct": ["recipient", "ccRecipient", "bccRecipient", "subject"],
+                        "slack-direct": [],
+                        "msft-teams-direct": ["webhookUrl", "message"],
+                    };
+                    let templateVariables = angular.copy(templateFields[$scope.agentTool.params.integration.type]);
+                    let defaultVariables = [];
+                    if ($scope.agentTool.params.integration.type === "mail-direct") {
+                        if ($scope.agentTool.params.integration.configuration.messageSource === "TEMPLATE_FILE") {
+                            defaultVariables = defaultVariables.concat($scope.fileTemplateVariables);
+                        } else {
+                            templateVariables.push("message");
+                        }
+                    } else if ($scope.agentTool.params.integration.type === "slack-direct") {
+
+                        if ($scope.agentTool.params.integration.configuration.mode === 'WEBHOOK') {
+                            templateVariables.push("webhookUrl");
+                        } else {
+                            templateVariables = templateVariables.concat(["channel", "iconEmoji", "iconUrl", "authorizationToken", "userName"]);
+                        }
+                        if ($scope.agentTool.params.integration.configuration.messageMode === "BLOCK_KIT_API") {
+                            templateVariables.push("blocks");
+                        } else {
+                            templateVariables.push("message");
+                        }
+                    }
+                    return new Set(
+                        templateVariables.map(function(field) {
+                            let value = $scope.agentTool.params.integration.configuration[field] || "";
+                            return extractVariables(value);
+                        }).reduce((a, b) => a.concat(b), defaultVariables)
+                    );
+                }
+
+                const getToolVariablesSet = function() {
+                    return new Set($scope.agentTool.params.variables.map(variable => variable.id));
+                }
+
+                const compareTemplateAndToolVariables = function() {
+                    let templateVariables = getTemplateVariablesSet();
+                    let toolVariables = getToolVariablesSet();
+                    $scope.undefinedTemplateVariables = Array.from(templateVariables.difference(toolVariables));
+                    $scope.unusedToolVariables = Array.from(toolVariables.difference(templateVariables));
+                };
+
+                $scope.addUndefinedVariables = function() {
+                    $scope.undefinedTemplateVariables.forEach((variable) => {
+                         $scope.agentTool.params.variables
+                        .push({
+                            "id": variable,
+                            "sourceType": "TOOL_INPUT",
+                        });
+                    });
+                    $scope.agentTool.params.variables = angular.copy($scope.agentTool.params.variables); // trigger UI update
+                };
+                $scope.removeUnusedVariables = function() {
+                    let unused = new Set($scope.unusedToolVariables);
+                    $scope.agentTool.params.variables = $scope.agentTool.params.variables.filter((variable) => !unused.has(variable.id));
+                };
+                $scope.isUnusedVariable = function(id) {
+                    return $scope.unusedToolVariables.includes(id);
+                }
+                $scope.hasUnusedVariable = function() {
+                    return $scope.unusedToolVariables.length > 0;
+                }
+                const makeDelayFn = function(fn, timerMs) {
+                    let promise = undefined;
+                    return function() {
+                        $timeout.cancel(promise);
+                        promise = $timeout(fn, timerMs || 0);
+                    };
+                }
+
+                const delayedCompareTemplate = makeDelayFn(compareTemplateAndToolVariables, 200);
+                $scope.$watchCollection('agentTool.params.integration.configuration', (newValue, oldValue) => {
+                    if (newValue.templateName !== oldValue.templateName) {
+                        loadEmailTemplate();
+                    } else {
+                        delayedCompareTemplate();
+                    };
+                });
+                $scope.variableChanged = delayedCompareTemplate;
+                $scope.$watch("agentTool.params.integration.type", compareTemplateAndToolVariables);
+                compareTemplateAndToolVariables();
+                loadEmailTemplate();
             }
 
             if ($scope.agentTool.type == "ClassicalPredictionModelPredict") {
@@ -304,7 +669,7 @@ app.controller('AgentToolController', function ($scope, $state, $stateParams, $q
                             if (endpoint) {
                                 $scope.agentTool.params.deploymentId = endpoint.deploymentId;
                                 $scope.agentTool.params.endpointId = endpoint.endpointId;
-                                $scope.agentTool.params.$description = endpoint.deploymentId;
+                                $scope.agentTool.params.$description = endpoint.description;
                                 $scope.uiState = {
                                     ...$scope.uiState,
                                     hasError: false,
@@ -462,7 +827,7 @@ app.controller('NewAgentToolController', function ($scope, $state, $stateParams,
         if (!$scope.newAgentTool.name) {
             $scope.newAgentTool.name = $scope.defaultName;
         }
-        const quickTestQuery = {"input": AgentToolService.getAgentToolQuickTestQueryForType($scope.selection.selectedObject.type), "context": {}};
+        const quickTestQuery = AgentToolService.getAgentToolQuickTestQueryForType($scope.selection.selectedObject.type);
         DataikuAPI.agentTools
             .create($stateParams.projectKey, { ...$scope.newAgentTool, type: $scope.selection.selectedObject.type, quickTestQuery: quickTestQuery })
             .success(function (data) {
@@ -493,12 +858,20 @@ app.component("agentToolQuicktest", {
         datasetSchema: "<",
     },
     templateUrl : '/templates/agent-tools/quicktest.html',
-    controller: function($scope, $timeout, DataikuAPI, AgentToolService, ClipboardUtils, CodeMirrorSettingService, WT1) {
+    controller: function($scope, $timeout, $q, DataikuAPI, AgentToolService, ClipboardUtils, CodeMirrorSettingService, SavedModelHelperService, TraceExplorerService, WT1) {
         const $ctrl = this;
         $ctrl.codeMirrorSettingService = CodeMirrorSettingService;
-        $ctrl.quickTestResponse = undefined;
+        $ctrl.quickTestResponse = null;
         $ctrl.quickTestEnabled = true;
+        $ctrl.quickTestError = null;
+        $ctrl.quickTestRequestCancellable = null;
+        $ctrl.stopDevKernelEnabled = true;
+        $ctrl.showStopDevKernelBtn = false;
+
         $scope.activeTab = 'response';
+        $scope.uiState = {
+            trace: 'FULL', // or GRAPHICAL
+        };
         $ctrl.hasCustomQuickTestQueryResolved;
 
         $ctrl.loadDescriptor = function() {
@@ -520,22 +893,42 @@ app.component("agentToolQuicktest", {
 
         };
 
+        $ctrl.validateQuicktestQuery = function() {
+            if (!$ctrl.quickTestQuery || typeof $ctrl.quickTestQuery !== 'object' || Array.isArray($ctrl.quickTestQuery)) {
+                $ctrl.quickTestError = "Invalid test query, it should be a JSON object";
+                return;
+            }
+            $ctrl.quickTestError = null;
+        }
+
         $ctrl.quickTest = function() {
             if (!$ctrl.quickTestEnabled) {
                 return;
             }
             $ctrl.quickTestEnabled = false;
+
             $ctrl.saveTool().then(() => {
-                DataikuAPI.agentTools.test($ctrl.projectKey, $ctrl.agentToolId, $ctrl.quickTestQuery).success((data) => {
+                $ctrl.stopDevKernelEnabled = true;
+                $ctrl.quickTestRequestCancellable = $q.defer();
+
+                DataikuAPI.agentTools.test($ctrl.projectKey, $ctrl.agentToolId, $ctrl.quickTestQuery, $ctrl.quickTestRequestCancellable.promise).then((response) => {
                     WT1.event(
                         'agent-tool-test-run', {
                             agentToolType: $ctrl.agentToolType,
                         });
-                    $ctrl.quickTestResponse = data;
+                    $ctrl.quickTestResponse = response.data;
+                    $ctrl.graphicalTrace = SavedModelHelperService.generateAsciiTreeFromTrace(response.data.fullTrace);
                     $scope.activeTab = 'response';
-                }).error(setErrorInScope.bind($scope)
-                ).finally(() => {
+                }).catch((error) => {
+                    if (error && error.status === -1) {
+                        // request was canceled
+                    } else {
+                        setErrorInScope.bind($scope)(error);
+                    }
+                    $ctrl.quickTestResponse = null;
+                }).finally(() => {
                     $ctrl.quickTestEnabled = true;
+                    $ctrl.quickTestRequestCancellable = null;
                 });
             });
         };
@@ -570,11 +963,33 @@ app.component("agentToolQuicktest", {
             });
         }
 
-        $ctrl.copyTraceToClipboard = function(value) {
+        $ctrl.stopDevKernel = function() {
+            if ($ctrl.quickTestRequestCancellable) {
+                $ctrl.quickTestRequestCancellable.resolve("Request canceled by user");
+            }
+            DataikuAPI.agentTools.stopDevKernel($ctrl.projectKey, $ctrl.agentToolId)
+                .then(() => {
+                    $ctrl.stopDevKernelEnabled = false;
+                })
+                .catch((error) => {
+                    if (error && error.status === 404) {
+                        $ctrl.stopDevKernelEnabled = false;
+                    } else {
+                        $ctrl.stopDevKernelEnabled = true;
+                        setErrorInScope.bind($scope)(error);
+                    }
+                });
+        }
+
+        $ctrl.copyTraceToClipboard = function() {
             $scope.traceCopied = true;
             ClipboardUtils.copyToClipboard(JSON.stringify($ctrl.quickTestResponse.fullTrace));
             $timeout(() => { $scope.traceCopied = false}, 5000);
         };
+        
+        $ctrl.openTraceExplorer = function() {
+            TraceExplorerService.openTraceExplorer($ctrl.quickTestResponse.fullTrace)
+        };
 
         $scope.$on('agentToolSavedEvent', () => {
             // only reload the descriptor if previously loaded manually
@@ -583,12 +998,18 @@ app.component("agentToolQuicktest", {
             }
         });
 
-        $scope.$on('agentToolSetEvent', () => {
-            if ($ctrl.agentToolType && !$ctrl.quickTestQuery) {
-                $ctrl.quickTestQuery = {"input": AgentToolService.getAgentToolQuickTestQueryForType($ctrl.agentToolType), "context": {}};
+        $ctrl.$onChanges = function (changes) {
+            if (changes.agentToolType) {
+                const newAgentToolType = changes.agentToolType.currentValue;
+                if (newAgentToolType) {
+                    $ctrl.showStopDevKernelBtn = AgentToolService.checkShowStopDevKernelBtn(newAgentToolType);
+                    if (!$ctrl.quickTestQuery) {
+                        $ctrl.quickTestQuery = AgentToolService.getAgentToolQuickTestQueryForType(newAgentToolType);
+                    }
+                    $ctrl.checkCustomQuickTestQuery();
+                }
             }
-            $ctrl.checkCustomQuickTestQuery();
-        });
+        };
     },
 });
 
@@ -653,7 +1074,7 @@ app.directive('agentToolRightColumnSummary', function($controller, $state, $stat
             scope.$on("objectSummaryEdited", save);
 
             scope.$watch("selection.selectedObject",function() {
-                if(scope.selection.selectedObject != scope.selection.confirmedItem) {
+                if (scope.selection.selectedObject != scope.selection.confirmedItem) {
                     scope.agentTool = null;
                     scope.objectTimeline = null;
                 }
@@ -873,6 +1294,7 @@ app.component("agentToolSelector", {
     bindings: {
         availableAgentTools: "<",
         selectedAgentTool: "=",
+        contextProject: "<",
         onChange: "<",
         defaultOpened: '<',
         disableAlertWarning: '<',
@@ -894,7 +1316,10 @@ app.component("agentToolSelector", {
                                   ng-select--custom-dropdown-height 
                                   ng-select--custom-x-overflow
                                   ng-select--custom-reset-default-padding"></basic-select>
-<div class="mbot0 mtop8 alert alert-warning" ng-if="!$ctrl.isAvailable && !$ctrl.disableAlertWarning">
+<div class="mbot0 mtop8 alert alert-warning" ng-if="$ctrl.selectedToolIsForeign()">
+Unable to display this tool properly: it was defined in the project {{$ctrl.getToolProjectKey()}} and not shared with the current project. Note that the tool or the project may not exist (e.g. if the current project was imported but the other project wasn't), or may have been deleted.
+</div>
+<div class="mbot0 mtop8 alert alert-warning" ng-if="!$ctrl.selectedToolIsForeign() && !$ctrl.isAvailable && !$ctrl.disableAlertWarning">
 Agent tool not available. It has been disabled, retired, or is not supported for this feature. Choose another Agent tool or contact your administrator. <strong>{{$ctrl.selectedAgentTool.id}}</strong>
 </div>`,
     controller: function ($scope, $filter) {
@@ -920,6 +1345,20 @@ Agent tool not available. It has been disabled, retired, or is not supported for
             $ctrl.updateAvailability();
         };
 
+        $ctrl.getToolProjectKey = function () {
+            if ($ctrl.selectedAgentTool && $ctrl.selectedAgentTool.includes(".")) {
+                return $ctrl.selectedAgentTool.split(".")[0];
+            }
+        };
+
+        $ctrl.selectedToolIsForeign = function () {
+            const toolProjectKey = $ctrl.getToolProjectKey();
+            if (toolProjectKey && toolProjectKey !== $ctrl.contextProject) {
+                return true;
+            }
+            return false;
+
+        };
         $ctrl.isAvailable = true;
         $ctrl.updateAvailability = () => {
             if ($ctrl.availableAgentTools && $ctrl.selectedAgentTool) {
