@@ -421,19 +421,21 @@ app.directive("objectGitHistory", function($stateParams, DataikuAPI) {
 
             $scope.$watch('objectId', function(nv) {
                 $scope.objectRef.id = nv;
+                $scope.loadMore();
             });
 
             $scope.loadMore = function () {
-                if ($scope.hasMore && !$scope.loading) {
+                const id = $scope.objectId || $stateParams.webAppId || $stateParams.smId;
+                if ($scope.hasMore && !$scope.loading && id) {
                     $scope.loading = true;
-                    DataikuAPI.git.getObjectLog($scope.projectKey || $stateParams.projectKey, $scope.objectType, $scope.objectId || $stateParams.webAppId || $stateParams.smId, $scope.nextCommit, PAGE_SIZE, $scope.projectZone).success(function(data){
+                    DataikuAPI.git.getObjectLog($scope.projectKey || $stateParams.projectKey, $scope.objectType, id, $scope.nextCommit, PAGE_SIZE, $scope.projectZone).success(function(data){
                         $scope.logEntries = ($scope.logEntries || []).concat(data.logEntries);
                         $scope.nextCommit = data.nextCommit;
                         if (!$scope.nextCommit) {
                             $scope.hasMore = false;
                         }
                         $scope.loading = false;
-                    }).error(function(e) {
+                    }).error(function() {
                         $scope.loading = false;
                         setErrorInScope.bind($scope);
                     });
