@@ -53,7 +53,7 @@
 
             const tooltips = ChartTooltips.create($rootElement, chartHandler, chartData, chartDef, tooltipFormatters);
             $container.on('click', function(evt) {
-                if (evt.target.hasAttribute('data-legend') || evt.target.hasAttribute('tooltip-el')) {
+                if (evt.target.closest('[data-legend], [tooltip-el]')) {
                     return;
                 }
                 tooltips.unfix();
@@ -82,14 +82,14 @@
             return agGridConverter.convert();
         }
 
-        function updateColumnVisibilities(gridOptions, chartDef, store) {
+        function updateColumnVisibilities(gridApi, chartDef, store) {
             const show = chartDef.pivotTableOptions.tableFormatting.showRowHeaders;
 
             chartDef.yDimension.forEach((dimension) => {
-                gridOptions.columnApi.setColumnVisible(store.getDimensionId(dimension), show);
+                gridApi.setColumnsVisible([store.getDimensionId(dimension)], show);
             });
 
-            gridOptions.columnApi.setColumnVisible(MEASURE_COLUMN_KEY, show);
+            gridApi.setColumnsVisible([MEASURE_COLUMN_KEY], show);
         }
 
         function showRowMainHeaderContextualMenu(event, cellValue, chartDef, mainHeader) {
@@ -181,10 +181,10 @@
          */
         function initGrid($container, gridOptions, chartDef, store) {
             // @ts-ignore
-            new AgGrid.Grid($container[0], gridOptions);
-            gridOptions.api.gridBodyCtrl.setCellTextSelection(true);
-            updateColumnVisibilities(gridOptions, chartDef, store);
-            chartDef.pivotTableOptions.showSidebar ? gridOptions.api.setSideBarVisible(true): gridOptions.api.setSideBarVisible(false);
+            const gridApi = AgGrid.createGrid($container[0], gridOptions);
+            store.setGridApi(gridApi);
+            updateColumnVisibilities(gridApi, chartDef, store);
+            chartDef.pivotTableOptions.showSidebar ? gridApi.setSideBarVisible(true): gridApi.setSideBarVisible(false);
             addTopHeadersContextualMenuListeners(chartDef);
         }
 
@@ -195,31 +195,29 @@
          */
         function updateGridOptions(newGridOptions, chartDef) {
             const store = ChartStoreFactory.get(chartDef.$chartStoreId);
-            const gridOptions = store.getGridOptions();
+            const gridApi = store.getGridApi();
+
 
             removeRowSubheadersContextualMenuListeners();
             updateTopHeadersContextualMenuListeners(chartDef);
 
-            Object.keys(newGridOptions).forEach(key => gridOptions[key] = newGridOptions[key]);
+            const { sideBar, columnDefs, ...newGridOptionsWithoutColumnDefsAndSideBar } = newGridOptions;
             if (chartDef.pivotTableOptions.showSidebar && typeof newGridOptions.sideBar !== 'string' && typeof newGridOptions.sideBar !== 'boolean') {
-                gridOptions.sideBar.toolPanels.forEach(toolpanel => {
-                    const instance = gridOptions.api.getToolPanelInstance(toolpanel.id);
-                    instance.customRefresh({ ...toolpanel.toolPanelParams, api: gridOptions.api, columnApi: gridOptions.columnApi });
+                gridApi.getSideBar().toolPanels.forEach((toolpanel, i) => {
+                    const instance = gridApi.getToolPanelInstance(toolpanel.id);
+                    instance.customRefresh({ ...sideBar.toolPanels[i].toolPanelParams, api: gridApi });
                 });
-                gridOptions.api.setSideBarVisible(true);
+                gridApi.setSideBarVisible(true);
             } else {
-                gridOptions.api.setSideBarVisible(false);
+                gridApi.setSideBarVisible(false);
             }
+            // If not it will merge the previous columnDefs with the new ones, leading to issues when removing measure
+            gridApi.setGridOption('columnDefs', columnDefs, { maintainColumnOrder: false });
 
-            gridOptions.api.setColumnDefs(newGridOptions.columnDefs);
-            gridOptions.api.setAutoGroupColumnDef(newGridOptions.autoGroupColumnDef);
-            gridOptions.api.setRowData(newGridOptions.rowData);
-            gridOptions.api.setGetRowHeight(() => newGridOptions.rowHeight);
-            gridOptions.api.resetRowHeights();
-            gridOptions.api.redrawRows();
-            gridOptions.api.refreshHeader();
-            gridOptions.columnApi.resetColumnState();
-            updateColumnVisibilities(gridOptions, chartDef, store);
+            gridApi.updateGridOptions(newGridOptionsWithoutColumnDefsAndSideBar);
+            gridApi.resetColumnState();
+
+            updateColumnVisibilities(gridApi, chartDef, store);
         }
 
         /**
@@ -249,15 +247,18 @@
          */
         function canRefreshView($element, chartDef, axesDef) {
             const store = ChartStoreFactory.get(chartDef.$chartStoreId);
-            const gridOptions = store.getGridOptions();
+            const gridApi = store.getGridApi();
+
+            if (gridApi == null) {
+                return false;
+            }
 
             /*
              * When changing the "Row headers" toggle from false to true for the first time, we need to switch to treeData mode, which means re-building the pivot from scratch.
              * This is because there is no way as of today in agGrid to hide the main group column so when hiding the row headers, we disable treeData mode entirely.
              */
-            const shouldRebuildTreeData = chartDef.pivotTableOptions.tableFormatting.showRowHeaders && (gridOptions || {}).treeData === false;
-            return gridOptions &&
-                !shouldRebuildTreeData &&
+            const shouldRebuildTreeData = chartDef.pivotTableOptions.tableFormatting.showRowHeaders && gridApi.getGridOption('treeData') === false;
+            return !shouldRebuildTreeData &&
                 !isNewRootElement($element)
                 && !hasChangedAxesDefinition(axesDef);
         }
@@ -268,16 +269,17 @@
          */
         function purgeAgGrid(chartDef) {
             const store = ChartStoreFactory.get(chartDef.$chartStoreId);
-            const gridOptions = store.getGridOptions();
+            const gridApi = store.getGridApi();
 
             if ($container && $container[0]) {
                 removeTopHeadersContextualMenuListeners();
                 removeRowSubheadersContextualMenuListeners();
             }
 
-            if (gridOptions && gridOptions.api) {
-                gridOptions.api.destroy();
+            if (gridApi != null && gridApi.destroy) {
+                gridApi.destroy();
             }
+            store.setGridApi(null);
         }
 
         /**
@@ -341,7 +343,6 @@
                 purgeAgGrid(chartDef);
                 $container = initContainer($element);
                 const gridOptions = generateGridOptions(chartDef, axesDef, data, chartHandler, $element, $container);
-                store.setGridOptions(gridOptions);
                 const requestOptions = store.getRequestOptions();
                 if (requestOptions && _.isNil(requestOptions.removePivotTableLimitations)) {
                     store.setRequestOptions({ removePivotTableLimitations: false });
