@@ -14,6 +14,9 @@
         this._init(element, options, draggableOptions, resizableOptions);
     };
 
+    // Minimum number of rows to add below the grid to allow dragging items to the bottom row when `addExtraGridSpaceBelow` is true
+    const MIN_EXTRA_GRID_SPACE = 10;
+
     const DragAndDropManager = {
         draggedItem: null,
         targetGrid: null,
@@ -1004,10 +1007,10 @@
              * right to allow dragging items to the end of the grid.
              */
             if (this.options.direction === 'horizontal') {
-                this.$element.width((this.gridList.grid.length + this.options.addExtraGridSpaceBelow ? this._widestItem : 0) * this._cellWidth + this.options.offsetWidth);
+                this.$element.width((this.gridList.grid.length + this.options.addExtraGridSpaceBelow ? Math.max(this._widestItem, MIN_EXTRA_GRID_SPACE) : 0) * this._cellWidth + this.options.offsetWidth);
             } else {
                 this.$element.height(
-                    (this.gridList.grid.length + (!this.options.readOnly && this.options.addExtraGridSpaceBelow ? this._tallestItem : 0)) * this._cellHeight + this.options.offsetHeight
+                    (this.gridList.grid.length + (!this.options.readOnly && this.options.addExtraGridSpaceBelow ? Math.max(this._tallestItem, MIN_EXTRA_GRID_SPACE) : 0)) * this._cellHeight + this.options.offsetHeight
                 );
             }
         },
